/*==========================================================================
	Datei:	CPU_SMA5D3_1.c

	fr FreeRTOS bentigte Programmierroutinen der Cortex A5 CPU SAMA5D3x.

	Begonnen:			06.02.2013
	Letzte nderung:	06.02.2013 (A. Heinze)
	Autor:				A.Heinze\HAMEG Chemnitz
===========================================================================*/

/*--- Includes ------------------------------------------------------------*/
#include <assert.h>
#include <limits.h>

#include <debug.h>
#include <IO_SAMA5D3x.h>

/*--- Makros & Konstanten -------------------------------------------------*/

/*--- Type Definitionen	---------------------------------------------------*/
typedef struct CPU_irq_s
{
	unsigned char	level;									//IRQ-Level vor Sperrung
	unsigned short	disableNo;								//Anzahl der Aufrufe von disable

#ifdef DEBUG_CPU_IRQ_DISABLE_CALLER
	const void		*disableCaller[16];						//Tabelle der letzten Aufrufer
	const void		*enableCaller[16];						//Tabelle der letzten Aufrufer
#endif
#ifdef DEBUG_CPU_IRQ_DISABLE_LOG_LENGTH
	unsigned int	logId;									// Index ins Log
	struct _cpu_irq_log
	{
		unsigned short	dis;								// 1 fr disable, 0 fr enable
		unsigned short	diNo;								// Zhlerstand von disableNo
		const void		*func;								// Aufrufer
	} log[DEBUG_CPU_IRQ_DISABLE_LOG_LENGTH];				// Log der letzten Aufrufer
#endif
} CPU_irq_type;

/*--- Globals -------------------------------------------------------------*/
CPU_irq_type	CPU_irq;
int				CPU_initIRQSystemDone = FALSE;

/*--- Prototypen ----------------------------------------------------------*/
// siehe _isrPort.s
extern unsigned int CPU_getCPSR(void);
extern void   CPU_initVBAR(void *);
extern void   CPU_resetVector(void *);

/*--- Funktionen ----------------------------------------------------------*/

void CPU_enableIRQ0(void)
{
#if CPU_IRQ_DISABLE_CALLER
	unsigned long lr;
	// Linkregister sichern
	asm(" mov	%0, lr" : "=r" (lr));
#endif

	//CPSR holen und auf FIQ Mode testen, wenn ja, dann raus hier
	if ((CPU_getCPSR() & CPU_CPSR_MODE_MASK) == CPU_CPSR_MODE_FIQ)
		return;

	// IRQs verbieten
	asm("	cpsid if");

#ifdef DEBUG_CPU_IRQ_DISABLE_LOG_LENGTH
	CPU_irq.log[CPU_irq.logId].dis = 0;
	CPU_irq.log[CPU_irq.logId].diNo = CPU_irq.disableNo;
	CPU_irq.log[CPU_irq.logId].func = (void *)lr;

#if (DEBUG_CPU_IRQ_DISABLE_LOG_LENGTH & (DEBUG_CPU_IRQ_DISABLE_LOG_LENGTH - 1)) == 0
	CPU_irq.logId = (CPU_irq.logId + 1) & (DEBUG_CPU_IRQ_DISABLE_LOG_LENGTH - 1);
#else
	CPU_irq.logId = (CPU_irq.logId + 1) % DEBUG_CPU_IRQ_DISABLE_LOG_LENGTH;
#endif
	CPU_irq.log[CPU_irq.logId].func = NULL;
#endif

	assert(CPU_irq.disableNo > 0);
	CPU_irq.disableNo--;

#ifdef DEBUG_CPU_IRQ_DISABLE_CALLER
	if (CPU_irq.disableNo < CALC_TAB_ELEMENTS(CPU_irq.enableCaller))
		CPU_irq.enableCaller[CPU_irq.disableNo] = (void *)lr;
#endif

	if ((CPU_irq.disableNo == 0) && (CPU_initIRQSystemDone != FALSE))
	{
		NDEBUG_PIN_LOW_(DEBUG_EDITOR_MBC, CPU_DEBUG_PIN_DISABLE_IRQ);

		//IRQs erlauben
		asm("	cpsie if");
	}
}	//CPU_enableIRQ0

unsigned int CPU_disableIRQ0(void)
{
#if CPU_IRQ_DISABLE_CALLER
	unsigned long lr;
	// Linkregister sichern
	asm(" mov	%0, lr" : "=r" (lr));
#endif

	//CPSR holen und auf FIQ Mode testen, wenn ja, dann raus hier
	if ((CPU_getCPSR() & CPU_CPSR_MODE_MASK) == CPU_CPSR_MODE_FIQ)
		return (0);

	// IRQs verbieten
	asm("	cpsid if");

#ifdef DEBUG_CPU_IRQ_DISABLE_CALLER
	if (CPU_irq.disableNo < CALC_TAB_ELEMENTS(CPU_irq.disableCaller))
		CPU_irq.disableCaller[CPU_irq.disableNo] = (void *)lr;
	if (CPU_irq.disableNo < CALC_TAB_ELEMENTS(CPU_irq.enableCaller))
		CPU_irq.enableCaller[CPU_irq.disableNo] = NULL;
#endif
#ifdef DEBUG_CPU_IRQ_DISABLE_LOG_LENGTH
	CPU_irq.log[CPU_irq.logId].dis = 1;
	CPU_irq.log[CPU_irq.logId].diNo = CPU_irq.disableNo;
	CPU_irq.log[CPU_irq.logId].func = (void *)lr;

#if (DEBUG_CPU_IRQ_DISABLE_LOG_LENGTH & (DEBUG_CPU_IRQ_DISABLE_LOG_LENGTH - 1)) == 0
	CPU_irq.logId = (CPU_irq.logId + 1) & (DEBUG_CPU_IRQ_DISABLE_LOG_LENGTH - 1);
#else
	CPU_irq.logId = (CPU_irq.logId + 1) % DEBUG_CPU_IRQ_DISABLE_LOG_LENGTH;
#endif
	CPU_irq.log[CPU_irq.logId].func = NULL;
#endif

	if (CPU_irq.disableNo == 0)
		NDEBUG_PIN_HIGH_(DEBUG_EDITOR_MBC, CPU_DEBUG_PIN_DISABLE_IRQ);

	assert(CPU_irq.disableNo < USHRT_MAX);
	CPU_irq.disableNo++;

	return (0);
}

void CPU_resetDisableIRQ (void)
{
	CPU_irq.disableNo = 0;						//Disable-Zhler

#ifdef DEBUG_CPU_IRQ_DISABLE_LOG_LENGTH
	CPU_irq.logId = 0;							//Log-ID
#endif
}

int CPU_readDisabledIRQ0(void)
{
	return (CPU_irq.disableNo != 0);
}

void CPU_setInitIRQSystemDone (int done)
{
	CPU_initIRQSystemDone = done;
}


int _cpu_checkIRQidx(unsigned int idx)
{
	if (idx >= CPU_IRQ_IDX_NO)
		return CPU_ERROR_WRONG_IRQ_INDEX;
	else
		return CPU_ERROR_NONE;
}

void CPU_setIRQIdx(unsigned int idx)
{
int error;

	error = _cpu_checkIRQidx(idx);
	if (error == CPU_ERROR_NONE)
	{
		CPU_disableIRQ0();
		//IRQ auswhlen
		REG_AIC_SSR = AIC_SSR_INTSEL(idx);
		//und Setzen
		REG_AIC_SMR |= AIC_SMR_SRCTYPE_INT_EDGE_TRIGGERED;
		REG_AIC_ISCR = AIC_ISCR_INTSET;
		CPU_enableIRQ0();
	}
}

